package italo.mlp;

public class MLP {
        
    /*
        Cria e inicializa a rede de acordo com a configuração. Após isso, se pode 
        iniciar o treinamento da rede.
    */
    public RNA criaRNA( RNAConfig config ) {                
        RNA rna = new RNA( config );
        
        int[] quantNeursPorCadaCamada = config.getQuantNeursPorCamada();
        int nCamadasOcultas = quantNeursPorCadaCamada.length-2;
        double wraio = config.getPesoIniRaio();
        double braio = config.getBiasIniRaio();
        
        double[][][] pesos = new double[ nCamadasOcultas + 1 ][][];
        double[][] biases = new double[ nCamadasOcultas + 1 ][];
        double[][] somas = new double[ nCamadasOcultas + 1 ][];
        double[][] deltas = new double[ nCamadasOcultas + 1 ][];
                
        double[][] saidas = new double[ nCamadasOcultas + 2 ][];
        int qAntNeus = quantNeursPorCadaCamada[0];
        
        // As entradas são também saídas das unidades da camada de entradas! 
        // Será inicializada no decorrer da propagação
        saidas[0] = new double[ quantNeursPorCadaCamada[0] ]; 
        
        for( int c = 1; c < nCamadasOcultas+2; c++ ) {
            int qNeurs = quantNeursPorCadaCamada[ c ];
            pesos[c-1] = new double[ qNeurs ][ qAntNeus ];
            biases[c-1] = new double[ qNeurs ];
            somas[c-1] = new double[ qNeurs ];
            deltas[c-1] = new double[ qNeurs ];
            saidas[c] = new double[ qNeurs ];            
            
            // inicialização dos pesos e biases
            for( int i = 0; i < qNeurs; i++ ) {                
                for( int j = 0; j < qAntNeus; j++ )
                    pesos[c-1][i][j] = -wraio*.5d + ( wraio * Math.random() );                                
                biases[c-1][i] = -braio*.5d + ( braio * Math.random() );
            }
            qAntNeus = qNeurs;
        }     
        
        rna.setPesos( pesos );
        rna.setBiases( biases );
        rna.setSomas( somas );
        rna.setSaidas( saidas );      
        rna.setDeltas( deltas ); 
        return rna;
    }    
    
    public void treinar( RNA rna ) {        
        int nEpocas = rna.getNEpocas();
        double[][] entradas = rna.getEntradas();
        double[][] saidasEsperadas = rna.getSaidasEsperadas();

        for( int epoca = 0; epoca < nEpocas; epoca++ ) {
            if ( epoca % 100 == 0 )
                System.out.println( "Epoca: "+epoca );
            int x_i = epoca % entradas.length;
                        
            double[] entrada = entradas[ x_i ];
            double[] saidaEsperada = saidasEsperadas[ x_i ];
            
            this.propaga( rna, entrada ); 
            this.retropropaga( rna, saidaEsperada );             
        }
    }
    
    /*             
       São várias unidades de entrada para uma entrada (vetor). Por exemplo: se a entrada 
       for 0 e 1, então, será uma unidade para representar a entrada 0 e outra para representar 
       a entrada 1. Cada unidade tem ligações sinápticas com os neurônios da camada seguinte!
    */
    public void propaga( RNADados rna, double[] entrada ) {
        double[][][] pesos = rna.getPesos();
        double[][] biases = rna.getBiases();
        double[][] somas = rna.getSomas();
        double[][] saidas = rna.getSaidas();
        
        saidas[0] = entrada;  // As entradas são também saídas das unidades da primeira camada, a camada de entradas!  
                
        for( int c = 0; c < pesos.length; c++ ) {                
            for( int j = 0; j < pesos[c].length; j++ ) {
                double u = 0;
                
                // cálculo dos potenciais de ativação (somas)
                for( int i = 0; i < pesos[c][j].length; i++ )
                    u += saidas[ c ][ i ] * pesos[c][j][i]; 
                    
                u += biases[ c ][ j ];
                                
                double s = 1.0d / ( 1.0d + Math.exp( -u ) ); // aplicação da função de ativação sigmoide

                somas[ c ][ j ] = u;
                saidas[ c+1 ][ j ] = s;                    
            }
        }
    }
    
    public void retropropaga( RNA rna, double[] saidaEsperada ) {
        double taxaAprendizado = rna.getPesosTaxaAprendizado();
        double[][][] pesos = rna.getPesos();
        double[][] biases = rna.getBiases();
        double[][] somas = rna.getSomas();
        double[][] saidas = rna.getSaidas();
        double[][] deltas = rna.getDeltas();
        
        for( int c = pesos.length-1; c >= 0; c-- ) {
            int j_n = rna.getNeursPorCamada()[ c+1 ];
            int i_n = rna.getNeursPorCamada()[ c ];                                
            for( int j = 0; j < j_n; j++ ) {                    
                double u = somas[c][j];
                double fatU = 1.0d / ( 1.0d + Math.exp( -u ) );
                
                double delta;
                if ( c == pesos.length-1 ) { // caso o índice j corresponda ao da última camada
                    delta = -2.0 * ( saidaEsperada[j] - saidas[c+1][j] ) * ( fatU * ( 1.0d - fatU ) );                                    
                } else { // caso o índice j corresponda a uma camada diferente da última, o delta local é acumulado
                    delta = 0;
                    int k_n = deltas[c+1].length;
                    for( int k = 0; k < k_n; k++ )
                        delta += deltas[c+1][k] * pesos[c+1][k][j];                    
                    delta *= ( fatU * ( 1.0d - fatU ) );
                }
                deltas[c][j] = delta;                                                                                           
                
                // atualização dos pesos
                for( int i = 0; i < i_n; i++ )                    
                    pesos[c][j][i] -= taxaAprendizado * delta * saidas[c][i];                  
                
                /* Perceba que o gradiente é o delta multiplicado pela saída anterior */                              
                
                // atualização dos biases
                biases[c][j] -= taxaAprendizado * delta;
            }
        }
    }
    
    /*
        Após a rede treinada, se pode utilizar esta função para o cálculo das saídas que, 
        após o treinamento, já devem estar bem próximas das desejadas.
    */
    public double[] calculaSaida( RNADados rnaDados, double[] entrada ) {        
        RNADados copia = rnaDados.copiaDados();
        this.propaga( copia, entrada );
        
        double[][][] pesos = copia.getPesos();
        double[][] saidas = copia.getSaidas();
        return saidas[ pesos.length ];        
    }
        
}
